module Doable
  # This class contains the actual work to be done
  class Step
    attr_reader :success, :status, :task, :name
    
    # @param options [Hash]
    # @param block [Proc]
    def initialize(context, options = {}, &block)
      @context  = context
      @success  = false
      @status   = :unattempted
      @task     = block
      @name     = options.key?(:name) ? options[:key] : block.to_s
      @rollback = nil
    end

    # Call our step, passing any arguments to the block (task) itself
    def call(*args)
      @status   = :failed     # first assume our attempt fails
      @context.instance_exec(*args, &@task)       # try to execute the Step's task
      @success  = true        # we'll only get here if no exceptions were raised
      @status   = :completed  # if we're here, the Step is complete
      return self
    end # call()
    
    # Set the Step's status to 'handled'
    def handled
      @status   = :handled
      @success  = true        # Might want to change this to false...
    end # handled

    # Boolean way of requesting success
    # @return [Boolean]
    def successful?
      return @success
    end
    
    # Wrapper around call used for retrying a step. Recommended approach to retrying as this may be
    # enhanced in the future.
    def retry(*args)
      call(*args)
    end

    # allow skipping steps (needs to be called externally)
    def skip
      @status   = :skipped
      @success  = false
    end

    # Sets up a block to call when rolling back this step
    # @return [Boolean]
    def rollback(&block)
      @rollback = block
      return true
    end

    # Query a step to see if it can be rolled back
    # @return [Boolean]
    def rollbackable?
      if @rollback and [:completed, :handled, :failed].include?(@status)
        return true
      else
        return false
      end
    end

    # Actually rollback this step
    def rollback!
      @context.instance_exec(&@rollback)
      @status = :rolledback
      @success = false
    end

  end # class Step
end